@@ -1,4 +1,4 @@
-#!/home/ben/software/install/bin/perl
+#!/usr/bin/env perl
 use warnings;
 use strict;
 use utf8;
@@ -11,6 +11,7 @@ package Monkey::Shines::Bool;
 sub true { my $monkey = 1; return bless \$monkey; }
 sub false { my $monkey = 0; return bless \$monkey; }
 1;
+package main;
 my $monkeys = {
     CuriousGeorge => Monkey::Shines->new (),
     KingKong => Monkey::Shines::Bool->true (),
@@ -26,7 +27,7 @@ my $obj_handler = sub {
 	return 'null';
     }
 };
-my $jc = JSON::Create->new ();
+my $jc = JSON::Create->new (indent => 1, sort => 1);
 print $jc->run ($monkeys), "\n";
 $jc->obj_handler ($obj_handler);
 print $jc->run ($monkeys), "\n";
