@@ -1,11 +1,11 @@
-#!/home/ben/software/install/bin/perl
+#!/usr/bin/env perl
 use warnings;
 use strict;
 use JSON::Create;
 my $jc = JSON::Create->new ();
 $jc->validate (1);
 $jc->type_handler (sub {
-		       return (1, 2, 3);
+		       return ('"a"', '"b"', '"c"');
 		   });
 print $jc->run ({ x => *STDOUT }); 
 
