@@ -1,5 +1,5 @@
-/* This file is a Unicode library in the programming language C which
-   deals with conversions to and from the UTF-8 format. */
+/* This is a Unicode library in the programming language C which deals
+   with conversions to and from the UTF-8 format. */
 
 /*
   Author: 
@@ -63,7 +63,7 @@
 #define UTF8_BAD_LEADING_BYTE -1
 
 /* This return value means the caller attempted to turn a code point
-   for a surrogate pair into UTF-8. */
+   for a surrogate pair to or from UTF-8. */
 
 #define UNICODE_SURROGATE_PAIR -2
 
@@ -88,7 +88,10 @@
 
 /* This return value indicates that UTF-8 bytes were not in the
    shortest possible form. See
-   http://www.cl.cam.ac.uk/~mgk25/unicode.html#utf-8. */
+   http://www.cl.cam.ac.uk/~mgk25/unicode.html#utf-8. 
+
+   This return value is currently unused. If a character is not in the
+   shortest form, the error UTF8_BAD_CONTINUATION_BYTE is returned. */
 
 #define UTF8_NON_SHORTEST -6
 
@@ -100,15 +103,17 @@
 
 /* This return value indicates that the Unicode code-point ended with
    either 0xFFFF or 0xFFFE, meaning it cannot be used as a character
-   code point. */
+   code point, or it was in the disallowed range FDD0 to FDEF. */
 
 #define UNICODE_NOT_CHARACTER -8
 
-/* This return value indicates that the UTF-8 is valid. */
+/* This return value indicates that the UTF-8 is valid. It is only
+   used by "valid_utf8". */
 
 #define UTF8_VALID 1
 
-/* This return value indicates that the UTF-8 is not valid. */
+/* This return value indicates that the UTF-8 is not valid. It is only
+   used by "valid_utf8". */
 
 #define UTF8_INVALID 0
 
@@ -165,6 +170,28 @@ int32_t utf8_bytes (uint8_t c)
     | (((int32_t) (x[2] & 0x3F)) <<  6)				\
     | (((int32_t) (x[3] & 0x3F)))
 
+/* Reject code points which end in either FFFE or FFFF. */
+
+#define REJECT_FFFF(x)				\
+    if ((x & 0xFFFF) >= 0xFFFE) {		\
+	return UNICODE_NOT_CHARACTER;		\
+    }
+
+/* Reject code points in a certain range. */
+
+#define REJECT_NOT_CHAR(r)					\
+    if (r >= UNI_NOT_CHAR_MIN && r <= UNI_NOT_CHAR_MAX) {	\
+	return UNICODE_NOT_CHARACTER;				\
+    }
+
+/* Reject surrogates. */
+
+#define REJECT_SURROGATE(ucs2)						\
+    if (ucs2 >= UNI_SUR_HIGH_START && ucs2 <= UNI_SUR_LOW_END) {	\
+	/* Ill-formed. */						\
+	return UNICODE_SURROGATE_PAIR;					\
+    }
+
 /* Try to convert "input" from UTF-8 to UCS-2, and return a value even
    if the input is partly broken.  This checks the first byte of the
    input, but it doesn't check the subsequent bytes. */
@@ -203,29 +230,59 @@ utf8_no_checks (const uint8_t * input, const uint8_t ** end_ptr)
     }
 }
 
-/* This function converts UTF-8 encoded bytes in "input" into the
-   equivalent Unicode code point. The return value is the Unicode code
-   point corresponding to the UTF-8 character in "input" if
-   successful, and a negative number if not successful. "*end_ptr" is
-   set to the next character after the read character on
-   success. "*end_ptr" is set to the start of input on
-   failure. "end_ptr" may not be null.
+/* Surrogate pair zone. */
 
-   If the first byte of "input" is zero, UNICODE_EMPTY_INPUT is
-   returned. If the first byte of "input" is not valid UTF-8,
-   UTF8_BAD_LEADING_BYTE is returned. If the second or later bytes of
-   "input" are not valid UTF-8, UTF8_BAD_CONTINUATION_BYTE is returned. If the
-   UTF-8 is not in the shortest possible form, the error
-   UTF8_NON_SHORTEST is returned. If the value extrapolated from
-   "input" is greater than UNICODE_MAXIMUM, UNICODE_TOO_BIG is
-   returned. If the value extrapolated from "input" ends in 0xFFFF or
-   0xFFFE, UNICODE_NOT_CHARACTER is returned. */
+#define UNI_SUR_HIGH_START  0xD800
+#define UNI_SUR_HIGH_END    0xDBFF
+#define UNI_SUR_LOW_START   0xDC00
+#define UNI_SUR_LOW_END     0xDFFF
+
+/* Start of the "not character" range. */
+
+#define UNI_NOT_CHAR_MIN    0xFDD0
+
+/* End of the "not character" range. */
+
+#define UNI_NOT_CHAR_MAX    0xFDEF
+
+/* This function converts UTF-8 encoded bytes in "input" into the
+   equivalent Unicode code point.  The return value is the Unicode
+   code point corresponding to the UTF-8 character in "input" if
+   successful, and a negative number if not successful.  Nul bytes are
+   rejected.
+
+   "*end_ptr" is set to the next character after the read character on
+   success.  "*end_ptr" is set to the start of input on all failures.
+   "end_ptr" may not be NULL.
+
+   If the first byte of "input" is zero, in other words a NUL or '\0',
+   UNICODE_EMPTY_INPUT is returned.
+
+   If the first byte of "input" is not valid UTF-8,
+   UTF8_BAD_LEADING_BYTE is returned.
+
+   If the second or later bytes of "input" are not valid UTF-8,
+   including NUL, UTF8_BAD_CONTINUATION_BYTE is returned.
+
+   If the value extrapolated from "input" is greater than
+   UNICODE_MAXIMUM, UNICODE_TOO_BIG is returned.
+
+   If the value extrapolated from "input" ends in 0xFFFF or 0xFFFE,
+   UNICODE_NOT_CHARACTER is returned.
+
+   If the value extrapolated from "input" is between 0xFDD0 and 0xFDEF, 
+   UNICODE_NOT_CHARACTER is returned.
+
+   If the value is within the range of surrogate pairs, the error
+   UNICODE_SURROGATE_PAIR is returned. 
+*/
 
 int32_t
 utf8_to_ucs2 (const uint8_t * input, const uint8_t ** end_ptr)
 {
     uint8_t c;
     uint8_t l;
+
     *end_ptr = input;
     c = input[0];
     if (c == 0) {
@@ -234,45 +291,57 @@ utf8_to_ucs2 (const uint8_t * input, const uint8_t ** end_ptr)
     l = utf8_sequence_len[c];
     if (l == 1) {
         * end_ptr = input + 1;
-        return c;
+        return (int32_t) c;
     }
     if (l == 2) {
+	uint8_t d;
+	d = input[1];
 	/* Two byte case. */
-        if (input[1] < 0x80 || input[1] > 0xBF) {
+        if (d < 0x80 || d > 0xBF) {
             return UTF8_BAD_CONTINUATION_BYTE;
 	}
 	if (c <= 0xC1) {
-	    return UTF8_NON_SHORTEST;
+            return UTF8_BAD_CONTINUATION_BYTE;
 	}
         * end_ptr = input + 2;
         return
             ((int32_t) (c & 0x1F) << 6)  |
-            ((int32_t) (input[1] & 0x3F));
+            ((int32_t) (d & 0x3F));
     }
     if (l == 3) {
+	uint8_t d;
+	uint8_t e;
+	int32_t r;
+
+	d = input[1];
+	e = input[2];
 	/* Three byte case. */
-        if (input[1] < 0x80 || input[1] > 0xBF ||
-	    input[2] < 0x80 || input[2] > 0xBF) {
+        if (d < 0x80 || d > 0xBF ||
+	    e < 0x80 || e > 0xBF) {
             return UTF8_BAD_CONTINUATION_BYTE;
 	}
-	if (c == 0xe0 && input[1] < 0xa0) { 
+	if (c == 0xe0 && d < 0xa0) { 
 	    /* We don't need to check the value of input[2], because
 	       the if statement above this one already guarantees that
 	       it is 10xxxxxx. */
-	    return UTF8_NON_SHORTEST;
+            return UTF8_BAD_CONTINUATION_BYTE;
 	}
+        r = ((int32_t) (c & 0x0F)) << 12 |
+            ((int32_t) (d & 0x3F)) << 6  |
+            ((int32_t) (e & 0x3F));
+	REJECT_SURROGATE(r);
+	REJECT_FFFF(r);
+	REJECT_NOT_CHAR(r);
         * end_ptr = input + 3;
-        return
-            ((int32_t) (c & 0x0F)) << 12 |
-            ((int32_t) (input[1] & 0x3F)) << 6  |
-            ((int32_t) (input[2] & 0x3F));
+	return r;
     }
-    if (l == 4) {
+    else if (l == 4) {
 	/* Four byte case. */
 	uint8_t d;
 	uint8_t e;
 	uint8_t f;
 	int32_t v;
+
 	d = input[1];
 	e = input[2];
 	f = input[3];
@@ -287,10 +356,10 @@ utf8_to_ucs2 (const uint8_t * input, const uint8_t ** end_ptr)
 	}
 
 	if (c == 0xf0 && d < 0x90) { 
-	    /* We don't need to check the values of e and d, because
+	    /* We don't need to check the values of e and f, because
 	       the if statement above this one already guarantees that
-	       e and d are 10xxxxxx. */
-	    return UTF8_NON_SHORTEST;
+	       e and f are 10xxxxxx. */
+            return UTF8_BAD_CONTINUATION_BYTE;
 	}
 	/* Calculate the code point. */
 	v = FOUR (input);
@@ -299,33 +368,40 @@ utf8_to_ucs2 (const uint8_t * input, const uint8_t ** end_ptr)
 	    return UNICODE_TOO_BIG;
 	}
 	/* Non-characters U+nFFFE..U+nFFFF on plane 1-16 */
-	if ((v & 0xffff) >= 0xfffe) {
-	    return UNICODE_NOT_CHARACTER;
-	}
+	REJECT_FFFF(v);
+	/* We don't need to check for surrogate pairs here, since the
+	   minimum value of UCS2 if there are four bytes of UTF-8 is
+	   0x10000. */
         * end_ptr = input + 4;
 	return v;
     }
     return UTF8_BAD_LEADING_BYTE;
 }
 
-#define UNI_SUR_HIGH_START  0xD800
-#define UNI_SUR_HIGH_END    0xDBFF
-#define UNI_SUR_LOW_START   0xDC00
-#define UNI_SUR_LOW_END     0xDFFF
 
 /* Input: a Unicode code point, "ucs2". 
 
    Output: UTF-8 characters in buffer "utf8". 
 
    Return value: the number of bytes written into "utf8", or a
-   negative number if there was an error. If the value of "ucs2" is
-   invalid because of being in the surrogate pair range from 0xD800 to
-   0xDFFF, the return value is UNICODE_SURROGATE_PAIR, else if the
-   value is too big to fit into four bytes of UTF-8, UNICODE_UTF8_4,
-   the return value is UNICODE_TOO_BIG. However, it does not insist on
-   ucs2 being less than UNICODE_MAXIMUM, so the user needs to check
-   that "ucs2" is a valid code point. It also does not check for
-   invalid characters, such as 0xFFFF.
+   negative number if there was an error. 
+
+   If the value of "ucs2" is invalid because of being in the surrogate
+   pair range from 0xD800 to 0xDFFF, the return value is
+   UNICODE_SURROGATE_PAIR.
+
+   If the value of "ucs2" is in the range 0xFDD0 to 0xFDEF inclusive,
+   the return value is UNICODE_NOT_CHARACTER.
+
+   If the lower two bytes of "ucs2" are either 0xFFFE or 0xFFFF, the
+   return value is UNICODE_NOT_CHARACTER.
+
+   If the value is too big to fit into four bytes of UTF-8,
+   UNICODE_UTF8_4, the return value is UNICODE_TOO_BIG.
+
+   However, it does not insist on ucs2 being less than
+   UNICODE_MAXIMUM, so the user needs to check that "ucs2" is a valid
+   code point.
 
    This adds a zero byte to the end of the string. It assumes that the
    buffer "utf8" has at least UNICODE_MAX_LENGTH (5) bytes of space to
@@ -334,6 +410,7 @@ utf8_to_ucs2 (const uint8_t * input, const uint8_t ** end_ptr)
 int32_t
 ucs2_to_utf8 (int32_t ucs2, uint8_t * utf8)
 {
+    REJECT_FFFF(ucs2);
     if (ucs2 < 0x80) {
         utf8[0] = ucs2;
         utf8[1] = '\0';
@@ -350,10 +427,8 @@ ucs2_to_utf8 (int32_t ucs2, uint8_t * utf8)
         utf8[1] = ((ucs2 >> 6 ) & 0x3F) | 0x80;
         utf8[2] = ((ucs2      ) & 0x3F) | 0x80;
         utf8[3] = '\0';
-	if (ucs2 >= UNI_SUR_HIGH_START && ucs2 <= UNI_SUR_LOW_END) {
-	    /* Ill-formed. */
-	    return UNICODE_SURROGATE_PAIR;
-	}
+	REJECT_SURROGATE(ucs2);
+	REJECT_NOT_CHAR(ucs2);
         return 3;
     }
     if (ucs2 <= UNICODE_UTF8_4) {
@@ -472,9 +547,9 @@ unicode_chars_to_bytes (const uint8_t * utf8, int32_t n_chars)
 }
 
 /* Like unicode_count_chars, but without error checks or validation of
-   the input. This only checks the first byte of each UTF-8
-   sequence. It may return UTF8_BAD_LEADING_BYTE if the first byte is
-   invalid. */
+   the input. This only checks the first byte of each UTF-8 sequence,
+   then jumps over the succeeding bytes. It may return
+   UTF8_BAD_LEADING_BYTE if the first byte is invalid. */
 
 int32_t
 unicode_count_chars_fast (const uint8_t * utf8)
@@ -562,6 +637,34 @@ unicode_count_chars (const uint8_t * utf8)
  case 0xB1: case 0xB2: case 0xB3: case 0xB4: case 0xB5: case 0xB6: case 0xB7: \
  case 0xB8: case 0xB9: case 0xBA: case 0xBB: case 0xBC: case 0xBD: case 0xBE: \
  case 0xBF
+#define BYTE_80_8F_B0_BF						\
+    0x80: case 0x81: case 0x82: case 0x83: case 0x84: case 0x85: case 0x86: \
+ case 0x87: case 0x88: case 0x89: case 0x8A: case 0x8B: case 0x8C: case 0x8D: \
+ case 0x8E: case 0x8F: case 0xB0:					\
+ case 0xB1: case 0xB2: case 0xB3: case 0xB4: case 0xB5: case 0xB6: case 0xB7: \
+ case 0xB8: case 0xB9: case 0xBA: case 0xBB: case 0xBC: case 0xBD: case 0xBE: \
+ case 0xBF
+#define BYTE_80_B6_B8_BF						\
+    0x80: case 0x81: case 0x82: case 0x83: case 0x84: case 0x85: case 0x86: \
+ case 0x87: case 0x88: case 0x89: case 0x8A: case 0x8B: case 0x8C: case 0x8D: \
+ case 0x8E: case 0x8F: case 0x90: case 0x91: case 0x92: case 0x93: case 0x94: \
+ case 0x95: case 0x96: case 0x97: case 0x98: case 0x99: case 0x9A: case 0x9B: \
+ case 0x9C: case 0x9D: case 0x9E: case 0x9F: case 0xA0: case 0xA1: case 0xA2: \
+ case 0xA3: case 0xA4: case 0xA5: case 0xA6: case 0xA7: case 0xA8: case 0xA9: \
+ case 0xAA: case 0xAB: case 0xAC: case 0xAD: case 0xAE: case 0xAF: case 0xB0: \
+ case 0xB1: case 0xB2: case 0xB3: case 0xB4: case 0xB5: case 0xB6: \
+ case 0xB8: case 0xB9: case 0xBA: case 0xBB: case 0xBC: case 0xBD: case 0xBE: \
+ case 0xBF
+#define BYTE_80_BD							\
+    0x80: case 0x81: case 0x82: case 0x83: case 0x84: case 0x85: case 0x86: \
+ case 0x87: case 0x88: case 0x89: case 0x8A: case 0x8B: case 0x8C: case 0x8D: \
+ case 0x8E: case 0x8F: case 0x90: case 0x91: case 0x92: case 0x93: case 0x94: \
+ case 0x95: case 0x96: case 0x97: case 0x98: case 0x99: case 0x9A: case 0x9B: \
+ case 0x9C: case 0x9D: case 0x9E: case 0x9F: case 0xA0: case 0xA1: case 0xA2: \
+ case 0xA3: case 0xA4: case 0xA5: case 0xA6: case 0xA7: case 0xA8: case 0xA9: \
+ case 0xAA: case 0xAB: case 0xAC: case 0xAD: case 0xAE: case 0xAF: case 0xB0: \
+ case 0xB1: case 0xB2: case 0xB3: case 0xB4: case 0xB5: case 0xB6: case 0xB7: \
+ case 0xB8: case 0xB9: case 0xBA: case 0xBB: case 0xBC: case 0xBD
 #define BYTE_90_BF							\
     0x90: case 0x91: case 0x92: case 0x93: case 0x94: case 0x95: case 0x96: \
  case 0x97: case 0x98: case 0x99: case 0x9A: case 0x9B: case 0x9C: case 0x9D: \
@@ -585,8 +688,6 @@ unicode_count_chars (const uint8_t * utf8)
 #define BYTE_E1_EC							\
     0xE1: case 0xE2: case 0xE3: case 0xE4: case 0xE5: case 0xE6: case 0xE7: \
  case 0xE8: case 0xE9: case 0xEA: case 0xEB: case 0xEC
-#define BYTE_EE_EF				\
-    0xEE: case 0xEF
 #define BYTE_F1_F3				\
     0xF1: case 0xF2: case 0xF3
 #endif /* def HEADER */
@@ -603,27 +704,76 @@ unicode_count_chars (const uint8_t * utf8)
 
 int32_t
 valid_utf8 (const uint8_t * input, int32_t input_length)
+{
+    int32_t error;
+    utf8_info_t info;
+    error = validate_utf8 (input, input_length, & info);
+    if (error < 0) {
+	return UTF8_INVALID;
+    }
+    return UTF8_VALID;
+}
+
+#define FAIL(x)				\
+    info->len_read = i;			\
+    return x
+
+#ifdef HEADER
+
+typedef struct utf8_info 
+{
+    int32_t len_read;
+    int32_t runes_read;
+}
+utf8_info_t;
+
+#endif /* def HEADER */
+
+/* Given "input" and "len", validate "input" byte by byte up to
+   "len". The return value is "UNICODE_OK" (zero) on success or the
+   error found (a negative number) on failure.
+
+   utf8_info_t is defined in "unicode.h".
+
+   The value of "info.len_read" is the number of bytes processed. the
+   value of "info.runes_read" is the number of Unicode code points in
+   the input. */
+
+int32_t
+validate_utf8 (const uint8_t * input, int32_t len, utf8_info_t * info)
 {
     int32_t i;
     uint8_t c;
 
+    info->len_read = 0;
+    /* We want to increment the runes after "string_start", but that
+       would give us one too many. */
+    info->runes_read = -1;
     i = 0;
 
  string_start:
 
-    i++;
-    if (i >= input_length) {
-	return UTF8_VALID;
+    /* We get here after successfully reading a "rune". */
+
+    info->runes_read++;
+    if (i >= len) {
+	info->len_read = len;
+	return UNICODE_OK; /* 0 */
     }
+
     /* Set c separately here since we use a range comparison before
        the switch statement. */
+
     c = input[i];
 
-    /* Admit all bytes <= 0x80. */
-    if (c <= 0x80) {
+    if (c == 0) {
+	FAIL (UNICODE_EMPTY_INPUT);
+    }
+    /* Admit all bytes < 0x80. */
+    if (c < 0x80) {
+	i++;
 	goto string_start;
     }
-
     switch (c) {
     case BYTE_C2_DF:
 	UNICODEADDBYTE;
@@ -641,10 +791,14 @@ valid_utf8 (const uint8_t * input, int32_t input_length)
 	UNICODEADDBYTE;
 	goto byte23_80_9f;
 	    
-    case BYTE_EE_EF:
+    case 0xEE:
 	UNICODEADDBYTE;
 	goto byte_penultimate_80_bf;
 	    
+    case 0xEF:
+	UNICODEADDBYTE;
+	goto byte_ef_80_bf;
+	    
     case 0xF0:
 	UNICODEADDBYTE;
 	goto byte24_90_bf;
@@ -657,86 +811,136 @@ valid_utf8 (const uint8_t * input, int32_t input_length)
 	UNICODEADDBYTE;
 	goto byte24_80_8f;
 
+    default:
+	FAIL (UTF8_BAD_LEADING_BYTE);
     }
 
  byte_last_80_bf:
 
     switch (UNICODENEXTBYTE) {
-
     case BYTE_80_BF:
 	UNICODEADDBYTE;
 	goto string_start;
     default:
-	UNICODEFAILUTF8 (XBYTES_80_BF);
+	FAIL (UTF8_BAD_CONTINUATION_BYTE);
+    }
+
+ byte_ef_b7:
+    switch (UNICODENEXTBYTE) {
+    case BYTE_80_8F_B0_BF:
+	UNICODEADDBYTE;
+	goto string_start;
+    default:
+	if (c >= 0x90 && c <= 0xAF) {
+	    FAIL (UNICODE_NOT_CHARACTER);
+	}
+	else {
+	    FAIL (UTF8_BAD_CONTINUATION_BYTE);
+	}
+    }
+
+ byte_last_80_bd:
+
+    switch (UNICODENEXTBYTE) {
+    case BYTE_80_BD:
+	UNICODEADDBYTE;
+	goto string_start;
+    case 0xBE:
+    case 0xBF:
+	FAIL (UNICODE_NOT_CHARACTER);
+    default:
+	FAIL (UTF8_BAD_CONTINUATION_BYTE);
     }
 
  byte_penultimate_80_bf:
 
     switch (UNICODENEXTBYTE) {
-
     case BYTE_80_BF:
 	UNICODEADDBYTE;
 	goto byte_last_80_bf;
     default:
-	UNICODEFAILUTF8 (XBYTES_80_BF);
+	FAIL (UTF8_BAD_CONTINUATION_BYTE);
+    }
+
+ byte_ef_80_bf:
+    switch (UNICODENEXTBYTE) {
+    case BYTE_80_B6_B8_BF:
+	UNICODEADDBYTE;
+	goto byte_last_80_bd;
+    case 0xB7:
+	UNICODEADDBYTE;
+	/* FDD0 - FDE7 */
+	goto byte_ef_b7;
+    default:
+	FAIL (UTF8_BAD_CONTINUATION_BYTE);
     }
 
  byte24_90_bf:
 
     switch (UNICODENEXTBYTE) {
-
     case BYTE_90_BF:
 	UNICODEADDBYTE;
 	goto byte_penultimate_80_bf;
     default:
-	UNICODEFAILUTF8 (XBYTES_90_BF);
+	FAIL (UTF8_BAD_CONTINUATION_BYTE);
     }
 
  byte23_80_9f:
 
     switch (UNICODENEXTBYTE) {
-
     case BYTE_80_9F:
 	UNICODEADDBYTE;
 	goto byte_last_80_bf;
     default:
-	UNICODEFAILUTF8 (XBYTES_80_9F);
+	if (c >= 0xA0 && c <= 0xBF) {
+	    FAIL (UNICODE_SURROGATE_PAIR);
+	}
+	else {
+	    FAIL (UTF8_BAD_CONTINUATION_BYTE);
+	}
     }
 
  byte23_a0_bf:
 
     switch (UNICODENEXTBYTE) {
-
     case BYTE_A0_BF:
 	UNICODEADDBYTE;
 	goto byte_last_80_bf;
     default:
-	UNICODEFAILUTF8 (XBYTES_A0_BF);
+	FAIL (UTF8_BAD_CONTINUATION_BYTE);
     }
 
  byte24_80_bf:
 
     switch (UNICODENEXTBYTE) {
-
     case BYTE_80_BF:
 	UNICODEADDBYTE;
-	goto byte_penultimate_80_bf;
+	goto byte_ef_80_bf;
     default:
-	UNICODEFAILUTF8 (XBYTES_80_BF);
+	FAIL (UTF8_BAD_CONTINUATION_BYTE);
     }
 
  byte24_80_8f:
 
     switch (UNICODENEXTBYTE) {
-
     case BYTE_80_8F:
 	UNICODEADDBYTE;
-	goto byte_penultimate_80_bf;
+	goto byte_ef_80_bf;
     default:
-	UNICODEFAILUTF8 (XBYTES_80_8F);
+	if (c >= 0x90) {
+	    FAIL (UNICODE_TOO_BIG);
+	}
+	else {
+	    FAIL (UTF8_BAD_CONTINUATION_BYTE);
+	}
     }
 }
 
+#define REJECT_FE_FF(c)				\
+    if (c == 0xFF || c == 0xFE) {		\
+	return UNICODE_NOT_CHARACTER;		\
+    }
+
 /* Make "* ptr" point to the start of the first UTF-8 character after
    its initial value. This assumes that there are at least four bytes
    which can be read, and that "* ptr" points to valid UTF-8. 
@@ -749,26 +953,35 @@ valid_utf8 (const uint8_t * input, int32_t input_length)
    second, third, or fourth byte of a multibyte sequence, "* ptr" is
    incremented until either "** ptr" is a valid first byte of a UTF-8
    sequence, or too many bytes have passed for it to be valid
-   UTF-8. If too many bytes have passed, UTF8_BAD_CONTINUATION_BYTE is returned
-   and "*ptr" is left unchanged. If a valid UTF-8 first byte was
-   found, either 11xx_xxxx or 00xx_xxxx, UNICODE_OK is returned, and
-   "*ptr" is set to the address of the valid byte. Nul bytes (bytes
-   containing zero) are considered valid. This does not check for
-   invalid UTF-8 bytes such as 0xFE and 0xFF. */
+   UTF-8. If too many bytes have passed, UTF8_BAD_CONTINUATION_BYTE is
+   returned and "*ptr" is left unchanged. 
+
+   If a valid UTF-8 first byte was found, either 11xx_xxxx or
+   00xx_xxxx, UNICODE_OK is returned, and "*ptr" is set to the address
+   of the valid byte. Nul bytes (bytes containing zero) are considered
+   valid.
+
+   If any of the bytes read contains invalid UTF-8 bytes 0xFE and
+   0xFF, the error code UNICODE_NOT_CHARACTER is returned and "*ptr"
+   is left unchanged. */
 
 int32_t
-trim_to_utf8_start (uint8_t ** ptr)
+trim_to_utf8_start (const uint8_t ** ptr)
 {
-    uint8_t * p = *ptr;
+    const uint8_t * p = *ptr;
     uint8_t c;
     int32_t i;
+
+    c = * p;
+    REJECT_FE_FF (c);
     /* 0xC0 = 1100_0000. */
-    c = *p & 0xC0;
+    c &= 0xC0;
     if (c == 0xC0 || c == 0x00) {
 	return UNICODE_OK;
     }
     for (i = 0; i < UTF8_MAX_LENGTH - 1; i++) {
 	c = p[i];
+	REJECT_FE_FF (c);
 	if ((c & 0x80) != 0x80 || (c & 0x40) != 0) {
 	    * ptr = p + i;
 	    return UNICODE_OK;
@@ -780,7 +993,8 @@ trim_to_utf8_start (uint8_t ** ptr)
 /* Given a return value "code" which is negative or zero, return a
    string which describes what the return value means. Positive
    non-zero return values never indicate errors or statuses in this
-   library. */
+   library. Unknown error codes result in a default string being
+   returned. */
 
 const char *
 unicode_code_to_error (int32_t code)
@@ -826,6 +1040,7 @@ unicode_code_to_error (int32_t code)
 #include "c-tap-test.h"
 
 static const uint8_t * utf8 = (uint8_t *) "漢数字ÔÕÖＸ";
+static const uint8_t bad[] = {0x99, 0x99, 0x99, 0x99, 0x99, 0x99, 0x0};
 
 #define BUFFSIZE 0x100
 
@@ -979,7 +1194,8 @@ test_utf8_to_ucs2 ()
         const uint8_t * end;
         unicode = utf8_to_ucs2 (start, & end);
 	TAP_TEST_MSG (unicode > 0, "no bad value at %s", start);
-        printf ("# %s is %04X, length is %d\n", start, unicode, end - start);
+        printf ("# %s is %04X, length is %d\n",
+		start, unicode, (int) (end - start));
         start = end;
     }
 }
@@ -1008,9 +1224,8 @@ static void
 test_trim_to_utf8_start ()
 {
     int32_t status;
-    uint8_t * p;
+    const uint8_t * p;
     /* Invalid UTF-8. */
-    uint8_t bad[] = {0x99, 0x99, 0x99, 0x99, 0x99, 0x99};
     /* Valid UTF-8. */
     uint8_t good[] = "化苦";
     uint8_t good2[] = "化abc";
@@ -1037,6 +1252,29 @@ test_constants ()
     TAP_TEST (UNICODE_UTF8_4 > UNICODE_MAXIMUM);
 }
 
+static void
+test_utf8_validate ()
+{
+    int r;
+    int l;
+    utf8_info_t info;
+
+    r = validate_utf8 ((const uint8_t *) "", 0, & info);
+    TAP_TEST_EQUAL (r, UNICODE_OK);
+    TAP_TEST_EQUAL (info.len_read, 0);
+    TAP_TEST_EQUAL (info.runes_read, 0);
+
+    l = strlen ((const char *) utf8);
+    r = validate_utf8 (utf8, l, & info);
+    TAP_TEST_EQUAL (r, UNICODE_OK);
+    TAP_TEST_EQUAL (info.len_read, l);
+    TAP_TEST_EQUAL (info.runes_read, 7);
+
+    l = strlen ((const char *) bad);
+    r = validate_utf8 (bad, l, & info);
+    TAP_TEST (r != UNICODE_OK);
+}
+
 int main ()
 {
     test_utf8_to_ucs2 ();
@@ -1048,6 +1286,7 @@ int main ()
     test_valid_utf8 ();
     test_trim_to_utf8_start ();
     test_constants ();
+    test_utf8_validate ();
     TAP_PLAN;
 }
 
